//
//  MWDocument.h
//  MySQL Workbench
//
//  Created by Alfredo Kojima on 05/6/17.
//  Copyright MySQL AB 2005 . All rights reserved.
//


#import <Cocoa/Cocoa.h>
#import "MWEditCanvasView.h"

#import <MySQLGRT/myx_grt_public_interface.h>
#import <MySQLGRT/MGRTShell.h>
#import <MySQLGRT/MGRT.h>
#import <MySQLGRT/MGRTTreeDataSource.h>

@class MTabView;

@class MWLayerDataSource;

class CLayer;

typedef enum {
  MWToolCursor = 0,
  MWToolPanning,
  MWToolEraser,
  MWToolTable,
  MWToolView,
  MWToolRoutine,
  MWToolRelationship1,
  MWToolRelationship2,
  MWToolRelationship3,
  MWToolImage,
  MWToolNote,
  MWToolLayer
} MWToolType;


@interface MWDocument : NSDocument
{
  IBOutlet NSWindow *modelWindow;
  IBOutlet NSView *paletteView;
  IBOutlet MTabView *tabView;
  IBOutlet NSDrawer *drawer;
  IBOutlet NSView *scrollTools;
  IBOutlet NSOutlineView *schemaTree;
  IBOutlet NSOutlineView *layerTree;
  IBOutlet NSPopUpButton *zoomPop;
  IBOutlet NSTextField *statusText;
  IBOutlet NSTextField *extraStatusText;
  
  IBOutlet NSView *layerOptions;
  IBOutlet NSView *tableOptions;
  IBOutlet NSView *cursorOptions;
  IBOutlet NSView *noteOptions;
  IBOutlet NSView *imageOptions;
  IBOutlet NSView *viewOptions;
  IBOutlet NSView *routineOptions;
  IBOutlet NSView *relationshipOptions;
  
  IBOutlet MGRTTreeDataSource *catalogDS;
  
  MCanvasScrollView *_scrollView;
  MWEditCanvasView *_canvas;

  MGRT *_grt;
  MGRTShell *_grtShell;
  
  NSArray *_toolCursors;
  NSView *_currentOptionsView;

  MWLayerDataSource *_layerDS;
  
  CLayer *_mainLayer;

  NSImage *_layerIcon;
  NSImage *_tableIcon;
  
  int _tableIndex;
  int _viewIndex;
  int _routineIndex;
  int _layerIndex;
  
  NSMutableDictionary *_openEditors;
  
  MYX_GRT_VALUE *_loadedDocumentData;
  
  MYX_GRT_VALUE *_relationshipContext;
  
  MWToolType _currentTool;
  BOOL _postedIdle;
}

- (IBAction)createSchema:(id)sender;

- (IBAction)toggleOverview:(id)sender;
- (IBAction)alignObjects:(id)sender;
- (IBAction)toggleGrid:(id)sender;
- (IBAction)toggleTool:(id)sender;
- (IBAction)changeZoom:(id)sender;
- (IBAction)pickImageFile:(id)sender;
- (IBAction)reverseEngineer:(id)sender;

- (void)updateToolOptions;
  
@end
