//
//  MWorkbench.mm
//  MySQL Workbench
//
//  Created by Alfredo Kojima on 05/8/22.
//  Copyright 2005 MySQL AB. All rights reserved.
//

#import "MWorkbench.h"
#define ENABLE_JAVA_MODULES
#import <MySQLGRT/MGRT.h>
#import <MySQLToolsCommon/MPreferences.h>
#import <WBCanvas/myx_gc.h>
#import <WBCanvas/myx_gc_font_manager.h>
#include <MySQLGRT/myx_grt_public_interface.h>

@implementation MWorkbench

- (void)applicationWillFinishLaunching:(NSNotification *)aNotification
{
  MYX_GRT *grt= myx_grt_initialize(0);
  MYX_GRT_MODULE_LOADER *loader;
  MYX_GRT_ERROR error;
  
  [[[NSThread currentThread] threadDictionary] setObject:[NSNumber numberWithInt:1] forKey:@"mainThread"];

  myx_grt_scan_for_structs(grt, [[NSString stringWithFormat:@"%@/xml",[[NSBundle bundleForClass:[MGRT class]] resourcePath]] fileSystemRepresentation],
                           &error);

  // create a GRT instance with a Java loader, so that the JavaVM will be first initialized
  // in the main thread. further GRT instances will just attach to the main JVM
  loader= myx_java_init_loader(grt, NULL, &error, NULL, 
                               [[NSString stringWithFormat:@"%@/", [[NSBundle bundleForClass:[MGRT class]] resourcePath]] fileSystemRepresentation]);
  if (loader)
  {
    myx_grt_register_module_loader(grt, loader);
    NSLog(@"java loader inited");
  }
  else
    NSLog(@"Error creating JVM!!");
  
  [MPreferences checkDirectory:@"Workbench"];
  
  // setup the global font manager with the font path
  lockFontManager();
  fontManager()->setFontResourcesPath([[[NSBundle mainBundle] resourcePath] UTF8String]);
}


- (IBAction)showAbout:(id)sender
{
  [versionLabel setStringValue:[NSString stringWithFormat:@"Version %@", 
    [[[NSBundle mainBundle] infoDictionary] objectForKey:@"CFBundleVersion"]]];
  
  [aboutPanel makeKeyAndOrderFront:sender];
}


@end
