unit ReverseEngineering;

interface

uses
  GnuGetText, Windows, Messages, SysUtils, Variants, Classes, Graphics,
  Controls, Forms, Dialogs,  TntForms, ExtCtrls, StdCtrls, TntStdCtrls,
  ComCtrls, TntComCtrls, TntExtCtrls,
  myx_public_interface, myx_grt_public_interface,
  MyxBaseForm, MyxConnectionDialog,
  PngImage, PngTools, Grt, ImgList, AdvancedEdit,
  MigrationObjSelFilter, AuxFuncs;

type
  TReverseEngineeringForm = class(TMyxBaseForm)
    MainPageControl: TTntPageControl;
    ConnectionSheet: TTntTabSheet;
    ConnParamGBox: TTntGroupBox;
    JdbcDriverDescLbl: TTntLabel;
    DriverParamHeaderLbl: TTntLabel;
    SourceTargetDBConnImg: TTntImage;
    ParamSBox: TTntScrollBox;
    DBConnAdvancedSettingsGBox: TTntGroupBox;
    AdvParamSBox: TTntScrollBox;
    SchemataSheet: TTntTabSheet;
    FinalizeSheet: TTntTabSheet;
    BottomPnl: TTntPanel;
    NextBtn: TTntButton;
    BackBtn: TTntButton;
    CancelBtn: TTntButton;
    HeaderPanel: TTntPanel;
    TntShape1: TTntShape;
    PageTitleLbl: TTntLabel;
    PageDescLbl: TTntLabel;
    GetSchemataSheet: TTntTabSheet;
    TntGroupBox1: TTntGroupBox;
    TntBevel1: TTntBevel;
    TntLabel3: TTntLabel;
    TntLabel7: TTntLabel;
    ConnImg: TTntImage;
    TntLabel8: TTntLabel;
    SchemaListImg: TTntImage;
    TntLabel4: TTntLabel;
    ResultLbl: TTntLabel;
    ProgressLbl: TTntLabel;
    AdvancedBtn: TTntButton;
    MessageLogGBox: TTntGroupBox;
    GRTMessageMemo: TTntMemo;
    RdbmsDriverGBox: TTntGroupBox;
    TntBevel2: TTntBevel;
    AnimPnl: TTntPanel;
    AnimStillImg: TTntImage;
    BusyAnimate: TAnimate;
    Shape1: TTntShape;
    TntGroupBox5: TTntGroupBox;
    SchemaSelectionCountLbl: TTntLabel;
    SchemaListView: TTntListView;
    SchemaSearchEd: TAdvancedEditFrame;
    ShowSchemataAsListCBox: TTntCheckBox;
    SmallSchemaListViewImgList: TImageList;
    SchemaListViewImgList: TImageList;
    RevEngSheet: TTntTabSheet;
    TntGroupBox2: TTntGroupBox;
    TntLabel5: TTntLabel;
    TntLabel6: TTntLabel;
    RevEngImg: TTntImage;
    TntLabel9: TTntLabel;
    CheckResImg: TTntImage;
    TntLabel10: TTntLabel;
    Result2Lbl: TTntLabel;
    Progress2Lbl: TTntLabel;
    RevEngMsgLogGBox: TTntGroupBox;
    GRTMessage2Memo: TTntMemo;
    IgnoreListSheet: TTntTabSheet;
    ObjSelScrollBox: TTntScrollBox;
    TntGroupBox4: TTntGroupBox;
    TntLabel15: TTntLabel;
    TntLabel16: TTntLabel;
    MigrationMappingSheet: TTntTabSheet;
    DefaultFramesScrollBox: TTntScrollBox;
    TntLabel11: TTntLabel;
    TntImage2: TTntImage;
    PlaceObjectsLbl: TTntLabel;
    PlaceObjectsImg: TTntImage;
    TntImage4: TTntImage;
    TntLabel1: TTntLabel;
    MigrationImg: TTntImage;
    MigrationLbl: TTntLabel;
    AddObjectsImg: TTntImage;
    AddObjectsLbl: TTntLabel;
    MessageLogFinalizeGBox: TTntGroupBox;
    GRTMessage3Memo: TTntMemo;
    ResultFinalizeLbl: TTntLabel;
    Progress3Lbl: TTntLabel;
    PlacementSheet: TTntTabSheet;
    TntGroupBox3: TTntGroupBox;
    TntLabel2: TTntLabel;
    TntLabel12: TTntLabel;
    PlaceAlgoComboBox: TTntComboBox;
    PlaceAlgoLbl: TTntLabel;
    PlaceAreaLbl: TTntLabel;
    PlaceAreaComboBox: TTntComboBox;
    PlaceSpaceLbl: TTntLabel;
    PlaceSpaceComboBox: TTntComboBox;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);

    procedure NextBtnClick(Sender: TObject);
    procedure BackBtnClick(Sender: TObject);
    procedure AdvancedBtnClick(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure SchemaSearchEdSearchEdChange(Sender: TObject);
    procedure ShowSchemataAsListCBoxClick(Sender: TObject);
    procedure SchemaListViewClick(Sender: TObject);
    procedure SchemaListViewSelectItem(Sender: TObject; Item: TListItem;
      Selected: Boolean);
    procedure FormResize(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure PlaceAlgoComboBoxCloseUp(Sender: TObject);
    procedure SchemaListViewDblClick(Sender: TObject);
  private
    FMyxConnDlgForm: TMyxConnectionDialogForm;

    FCancelOperation: Boolean;

    FConnection,
    FSchemaList,
    FCatalog: Pointer;

    FBackBtnEnabled,
    FNextBtnEnabled: Boolean;

    FTaskUncheckedPNGImg,
    FTaskCheckedPNGImg,
    FTaskErrorPNGImg,
    FTaskDisabledPNGImg: TPngObject;

    FNormalHeight,
    FAdvancedHeight: Integer;

    FFilterFrames: TList;

    // Task management
    FFetchSchemaTask: IGrtTask;
    FReverseEngineerTask: IGrtTask;
    FSelectedSchemaList: Pointer;
    FProcessCatalogsTask: IGrtTask;
    FArrangeTask: IGrtTask;
  protected
    procedure UpdateAdvancedBtnState;
    procedure UpdateSchemaSelectionCountLbl;
    procedure UpdatePageHeader;

    procedure DoFetchSchemaNames;
    procedure DoDisplaySchemata;
    procedure DoRevEng;
    procedure DoBuildObjectSelection;
    procedure DoFinalizeRevEng;

    procedure ProcessGrtOutput(Text: WideString);
    procedure ProcessGrtMessages(Messages: TMYX_GRT_MSGS);
    function ProcessStatusQuery: Integer;

    procedure RefreshSchemaListView(SearchStr: WideString = '');

    procedure DoProcessStart;
    procedure DoProcessEnd;

    function AdvancedSectionsVisible: Boolean;

    function PerformGrtInput(Description: WideString): WideString;

    procedure TaskFinished(Task: IGrtGenericTask);
    function TaskHandleError(Task: IGrtGenericTask): Boolean;
  public
  end;

implementation

{$R *.dfm}

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.FormCreate(Sender: TObject);

var
  Migration: Pointer;

begin
  FSchemaList := nil;
  FCatalog := nil;

  FNormalHeight := 512 + 80;
  FAdvancedHeight := 686 + 80;

  LoadPNGImageFromResource('sakila', AnimStillImg, True);
  BusyAnimate.ResName:='dolphin_loop';

  LoadPNGImageFromResource('source_dbconn', SourceTargetDBConnImg, True);

  FTaskUncheckedPNGImg := LoadPNGImageFromResource('task_unchecked');
  FTaskCheckedPNGImg := LoadPNGImageFromResource('task_checked');
  FTaskErrorPNGImg := LoadPNGImageFromResource('task_error');
  FTaskDisabledPNGImg := LoadPNGImageFromResource('task_disabled');

  FCancelOperation := False;

  FFilterFrames := TList.Create;

  MainPageControl.ActivePage := ConnectionSheet;

  // Dock connection form
  FMyxConnDlgForm := TMyxConnectionDialogForm.Create(nil);

  FMyxConnDlgForm.ConnInfoPath := '/rdbmsMgmt';
  FMyxConnDlgForm.ConnTargetPath :='/workbench/connection';

  FMyxConnDlgForm.DisplayRdbmsSelection := True;
  FMyxConnDlgForm.DisplaySchemaSelection := False;
  FMyxConnDlgForm.DisplayDescriptions := True;

  FMyxConnDlgForm.ConnTypePnl.Parent := RdbmsDriverGBox;
  FMyxConnDlgForm.ConnTypePnl.Left := 28;
  FMyxConnDlgForm.ConnTypePnl.Width := RdbmsDriverGBox.Width - 40;

  FMyxConnDlgForm.StoredConnPnl.Parent := ParamSBox;
  FMyxConnDlgForm.StoredConnPnl.Top := 0;

  FMyxConnDlgForm.ParamsPnl.Parent := ParamSBox;
  FMyxConnDlgForm.ParamsPnl.Top := FMyxConnDlgForm.StoredConnPnl.Height;
  FMyxConnDlgForm.ParamsPnl.Width := ParamSBox.Width - 20;

  FMyxConnDlgForm.AdvParamPnl.Parent := AdvParamSBox;
  FMyxConnDlgForm.AdvParamPnl.Top := 0;
  FMyxConnDlgForm.AdvParamPnl.Width := AdvParamSBox.Width - 30;

  FMyxConnDlgForm.DriverNotInstalledPnl.Parent := ParamSBox;
  FMyxConnDlgForm.DriverNotInstalledPnl.Top := 0;

  FMyxConnDlgForm.OkButton := NextBtn;

  FMyxConnDlgForm.RefreshConnInfo;

  FMyxConnDlgForm.SelectedRdbms := 'Mysql';

  Height := FNormalHeight;

  // Create migration object and add it to the globals tree
  Migration := Grt.ObjectNew('db.migration.Migration',
    'Migration', '', '');
  Grt.Global['/migration'] := Migration;
  Grt.ValueRelease(Migration);

  PlaceAlgoComboBox.ItemIndex := 1;
  PlaceAreaComboBox.ItemIndex := 1;
  PlaceSpaceComboBox.ItemIndex := 1;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.FormDestroy(Sender: TObject);

begin
  FTaskUncheckedPNGImg.Free;
  FTaskCheckedPNGImg.Free;
  FTaskErrorPNGImg.Free;
  FTaskDisabledPNGImg.Free;

  FFilterFrames.Free;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  Grt.GlobalDel('/', 'migration');

  if (FSchemaList <> nil) then
    Grt.ValueRelease(FSchemaList);

  if (FCatalog <> nil) then
    Grt.ValueRelease(FCatalog);
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.NextBtnClick(Sender: TObject);

begin
  BackBtn.Enabled := True;

  if (MainPageControl.ActivePage = ConnectionSheet) then
    DoFetchSchemaNames
  else
    if (MainPageControl.ActivePage = GetSchemataSheet) then
      DoDisplaySchemata
    else
      if (MainPageControl.ActivePage = SchemataSheet) then
        DoRevEng
      else
        if (MainPageControl.ActivePage = RevEngSheet) then
          DoBuildObjectSelection
        else
          if (MainPageControl.ActivePage = IgnoreListSheet) then
            MainPageControl.ActivePage := PlacementSheet
          else
            if (MainPageControl.ActivePage = PlacementSheet) then
              DoFinalizeRevEng
            else
              if (MainPageControl.ActivePage = FinalizeSheet) then
                Close;

  UpdatePageHeader;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.BackBtnClick(Sender: TObject);

begin
  if (MainPageControl.ActivePageIndex > 0) then
    MainPageControl.ActivePageIndex :=
      MainPageControl.ActivePageIndex - 1;

  if (MainPageControl.ActivePageIndex = 0) then
    BackBtn.Enabled := False;

  NextBtn.Enabled := True;

  UpdateAdvancedBtnState;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.CancelBtnClick(Sender: TObject);

begin
  FCancelOperation := True;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.AdvancedBtnClick(Sender: TObject);

begin
  if (MainPageControl.ActivePage = ConnectionSheet) then
    DBConnAdvancedSettingsGBox.Visible :=
      Not(DBConnAdvancedSettingsGBox.Visible)
  else
    if (MainPageControl.ActivePage = GetSchemataSheet) then
      MessageLogGBox.Visible :=
        Not(MessageLogGBox.Visible)
    else
      if (MainPageControl.ActivePage = RevEngSheet) then
        RevEngMsgLogGBox.Visible :=
          Not(RevEngMsgLogGBox.Visible)
      else
        if (MainPageControl.ActivePage = FinalizeSheet) then
          MessageLogFinalizeGBox.Visible :=
            Not(MessageLogFinalizeGBox.Visible);

  UpdateAdvancedBtnState;
end;

// -----------------------------------------------------------------------------

function TReverseEngineeringForm.AdvancedSectionsVisible: Boolean;

begin
  case MainPageControl.ActivePageIndex of
    0:
      Result := DBConnAdvancedSettingsGBox.Visible;
    1:
      Result := MessageLogGBox.Visible;
    3:
      Result := RevEngMsgLogGBox.Visible;
  else
    Result := False;
  end;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.UpdateAdvancedBtnState;

begin
  AdvancedBtn.Enabled := (MainPageControl.ActivePage <> SchemataSheet);

  if (AdvancedSectionsVisible) then
  begin
    AdvancedBtn.Caption := _('<< Advanced');

    Height := FAdvancedHeight;

    DBConnAdvancedSettingsGBox.Height := Height - (686 - 161);
    DBConnAdvancedSettingsGBox.Anchors :=
      [akLeft, akTop, akRight, akBottom];
  end
  else
  begin
    AdvancedBtn.Caption := _('Advanced >>');

    Height := FNormalHeight;
  end;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.UpdatePageHeader;

begin
  if (MainPageControl.ActivePage = ConnectionSheet) then
  begin
    PageTitleLbl.Caption := _('Source Connection');
    PageDescLbl.Caption := _('Please specify the source connection.');
  end
  else
    if (MainPageControl.ActivePage = GetSchemataSheet) then
    begin
      PageTitleLbl.Caption := _('Source Schemata');
      PageDescLbl.Caption := _('The list of available schemata is fetched.');
    end
    else
      if (MainPageControl.ActivePage = SchemataSheet) then
      begin
        PageTitleLbl.Caption := _('Schema Selection');
        PageDescLbl.Caption := _('Please select the schema(ta) to reverse engineer.');
      end
      else
        if (MainPageControl.ActivePage = RevEngSheet) then
        begin
          PageTitleLbl.Caption := _('Reverse Engineering');
          PageDescLbl.Caption := _('The selected schema(ta) are reverse engineered.');
        end
        else
          if (MainPageControl.ActivePage = IgnoreListSheet) then
          begin
            PageTitleLbl.Caption := _('Object Selection');
            PageDescLbl.Caption := _('Add objects to the ignore list.');
          end
          else
            if (MainPageControl.ActivePage = MigrationMappingSheet) then
            begin
              PageTitleLbl.Caption := _('Migration Mappings');
              PageDescLbl.Caption := _('Please select the methods that will be used to migrate the source objects.');
            end
            else
              if (MainPageControl.ActivePage = FinalizeSheet) then
              begin
                PageTitleLbl.Caption := _('Finalization');
                PageDescLbl.Caption := _('Finalizing the reverse engineering process.');
              end;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.ProcessGrtOutput(Text: WideString);

begin
  if (MainPageControl.ActivePage = GetSchemataSheet) then
    GRTMessageMemo.Lines.Add(Text)
  else
    if (MainPageControl.ActivePage = RevEngSheet) then
      GRTMessage2Memo.Lines.Add(Text)
    else
      if (MainPageControl.ActivePage = FinalizeSheet) then
      GRTMessage3Memo.Lines.Add(Text)
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.ProcessGrtMessages(
  Messages: TMYX_GRT_MSGS);

begin
  // display last message in ResultLbl
  if (Messages.msgs.Count > 0) then
  begin
    if (MainPageControl.ActivePage = GetSchemataSheet) then
    begin
      ProgressLbl.Caption := Messages.msgs[Messages.msgs.Count - 1].msg;
      ProgressLbl.Update;

      GRTMessageMemo.Text := GRTMessageMemo.Text +
        FormatGrtMessagesAsString(Messages);
      GRTMessageMemo.Update;
    end
    else
      if (MainPageControl.ActivePage = RevEngSheet) then
      begin
        Progress2Lbl.Caption := Messages.msgs[Messages.msgs.Count - 1].msg;
        Progress2Lbl.Update;

        GRTMessage2Memo.Text := GRTMessage2Memo.Text +
          FormatGrtMessagesAsString(Messages);
        GRTMessage2Memo.Update;
      end
      else
        if (MainPageControl.ActivePage = FinalizeSheet) then
        begin
          Progress3Lbl.Caption := Messages.msgs[Messages.msgs.Count - 1].msg;
          Progress3Lbl.Update;

          GRTMessage3Memo.Text := GRTMessage3Memo.Text +
            FormatGrtMessagesAsString(Messages);
          GRTMessage3Memo.Update;
        end;
  end;
end;

// -----------------------------------------------------------------------------

function TReverseEngineeringForm.ProcessStatusQuery: Integer;

begin
  Result := Ord(FCancelOperation);
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.SchemaSearchEdSearchEdChange(
  Sender: TObject);

begin
  SchemaSearchEd.SearchEdChange(Sender);

  RefreshSchemaListView(SchemaSearchEd.SearchEd.Text);
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.ShowSchemataAsListCBoxClick(
  Sender: TObject);
begin
  if (ShowSchemataAsListCBox.Checked) then
    SchemaListView.ViewStyle:=vsList
  else
    SchemaListView.ViewStyle:=vsIcon;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.SchemaListViewClick(Sender: TObject);

begin
  UpdateSchemaSelectionCountLbl;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.UpdateSchemaSelectionCountLbl;

begin
  if (SchemaListView.SelCount = 0) then
  begin
    SchemaSelectionCountLbl.Caption :=
      _('No schemata selected.');

    NextBtn.Enabled := False;
  end
  else
  begin
    if(SchemaListView.SelCount=1)then
      SchemaSelectionCountLbl.Caption :=
        _('1 schema selected.')
    else
      SchemaSelectionCountLbl.Caption :=
        Format(_('%d schemata selected.'), [SchemaListView.SelCount]);

    NextBtn.Enabled := True;
  end;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.RefreshSchemaListView(SearchStr: WideString);

var
  I: integer;
  ListItem: TTntListItem;
  S: WideString;

begin
  SchemaListView.Clear;

  if (FSchemaList <> nil) then
  begin
    for I:=0 to Grt.ListCount(FSchemaList) - 1 do
    begin
      S := Grt.ListString[FSchemaList, I];

      if (SearchStr = '') or
        (myx_match_pattern(S, SearchStr + '*', 0, 1) = 1) then
      begin
        ListItem := SchemaListView.Items.Add;
        ListItem.Caption := S;
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.SchemaListViewSelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);

begin
  UpdateSchemaSelectionCountLbl;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.DoProcessStart;

begin
  BusyAnimate.Show;
  BusyAnimate.Play(1, 74, -1);
  BusyAnimate.Active:=True;

  FBackBtnEnabled := BackBtn.Enabled;
  FNextBtnEnabled := NextBtn.Enabled;

  FCancelOperation := False;

  //WizardBottomFrame.DetailsBtn.Enabled := False;
  BackBtn.Enabled := False;
  NextBtn.Enabled := False;
  CancelBtn.Enabled := True;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.DoProcessEnd;

begin
  BusyAnimate.Hide;
  BusyAnimate.Stop;

  //WizardBottomFrame.DetailsBtn.Enabled := FDetailsBtnEnabled;
  BackBtn.Enabled := FBackBtnEnabled;
  NextBtn.Enabled := FNextBtnEnabled;
  CancelBtn.Enabled := False;

  FCancelOperation := False;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.FormResize(Sender: TObject);

begin
  if (AdvancedSectionsVisible) then
    FAdvancedHeight := Height
  else
    FNormalHeight := Height;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.DoFetchSchemaNames;

var
  RevEngModuleName: WideString;

begin
  FConnection := FMyxConnDlgForm.WriteConnectionToTarget;

  RevEngModuleName := Grt.DictString[Grt.DictItem[FConnection, 'modules'], 'ReverseEngineeringModule'];

  ConnImg.Picture.Assign(FTaskUncheckedPNGImg);
  SchemaListImg.Picture.Assign(FTaskUncheckedPNGImg);

  ResultLbl.Hide;
  GRTMessageMemo.Clear;
  ProgressLbl.Caption := '';
  ProgressLbl.Show;

  // Flip page
  MainPageControl.ActivePage := GetSchemataSheet;
  UpdateAdvancedBtnState;

  DoProcessStart;

  Application.ProcessMessages;
  // if there already is a FSchemaList, release it
  if (FSchemaList <> nil) then
    Grt.ValueRelease(FSchemaList);

  FFetchSchemaTask := Grt.CreateStandardTask('Getting schemata list', RevEngModuleName, 'getSchemata',
    [Grt.GetGlobalAsParam('/workbench/connection')], ProcessGrtOutput, ProcessGrtMessages, False, True, -1, nil,
    ProcessStatusQuery, TaskFinished);
  Grt.AddTask(FFetchSchemaTask);
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.DoDisplaySchemata;

begin
  RefreshSchemaListView;

  NextBtn.Enabled := False;

  // Flip page
  MainPageControl.ActivePage := SchemataSheet;
  UpdateAdvancedBtnState;

  try
    if(SchemaListView.CanFocus)then
      SchemaListView.SetFocus;
  except
  end;
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.DoRevEng;

var
  RevEngModuleName: WideString;
  I: Integer;

begin
  RevEngModuleName := Grt.DictString[Grt.DictItem[FConnection, 'modules'], 'ReverseEngineeringModule'];

  RevEngImg.Picture.Assign(FTaskUncheckedPNGImg);
  CheckResImg.Picture.Assign(FTaskUncheckedPNGImg);

  Result2Lbl.Hide;
  GRTMessageMemo.Clear;
  Progress2Lbl.Caption := '';
  Progress2Lbl.Show;

  // Flip page
  MainPageControl.ActivePage := RevEngSheet;
  UpdateAdvancedBtnState;

  DoProcessStart;

  Application.ProcessMessages;

  FSelectedSchemaList := Grt.ListNew(GrtStringValue);
  for I := 0 to SchemaListView.Items.Count - 1 do
    if (SchemaListView.Items[I].Selected) then
      Grt.ListAdd(FSelectedSchemaList, Grt.ValueFromString(SchemaListView.Items[I].Caption), False);

  if (FCatalog <> nil) then
    Grt.ValueRelease(FCatalog);

  FReverseEngineerTask := Grt.CreateStandardTask(
    'Reverse engineering schema(ta)', RevEngModuleName, 'reverseEngineer',
    [Grt.GetGlobalAsParam('/workbench/connection'), FSelectedSchemaList],
    ProcessGrtOutput, ProcessGrtMessages, False,
    True, -1, nil, ProcessStatusQuery, TaskFinished
  );
  Grt.AddTask(FReverseEngineerTask);
end;

// -----------------------------------------------------------------------------

procedure TReverseEngineeringForm.DoBuildObjectSelection;

var
  I, J, K: Integer;
  PSchemata, PSchema: Pointer;
  PSchemaStruct, PSchemaStructMember: Pointer;
  PSchemaItemList,
    PSchemaItem: Pointer;
  MemberName,
    ContentStructName: WideString;
  FilterFrame: TMigrationObjSelFilterFrame;
  PSourceObjects: Pointer;
  C: Integer;

begin
  //Clear Filter Frames if there are some already
  for I:=0 to FFilterFrames.Count-1 do
    TMigrationObjSelFilterFrame(FFilterFrames[I]).Free;
  FFilterFrames.Clear;

  PSchemata := Grt.DictItem[FCatalog, 'schemata'];

  //Get the struct type of the schemata
  PSchema := myx_grt_list_item_get(PSchemata, 0);
  PSchemaStruct := myx_grt_dict_struct_get(Grt.PGrt, PSchema);

  //Clear the Mig.SourceObjects list
  PSourceObjects := Grt.Global['/migration/sourceObjects'];
  myx_grt_list_clear(PSourceObjects);

  //Create a filter frame for each member deriving from db.DatabaseObject
  for I:=myx_grt_struct_get_member_count_total(Grt.PGrt, PSchemaStruct)-1 downto 0 do
  begin
    PSchemaStructMember :=
      myx_grt_struct_get_member_by_index_total(Grt.PGrt, PSchemaStruct, I);

    // if the member is not a list, continue
    if (Grt.GetStructMemberType(PSchemaStructMember) <> GrtListValue) then
      Continue;

    MemberName :=
      UTF8Decode(_myx_grt_struct_get_member_name(PSchemaStructMember));

    ContentStructName :=
      UTF8Decode(_myx_grt_struct_member_get_content_struct_name(PSchemaStructMember));

    //Only consider structs that derive from db.DatabaseObject
    if (myx_grt_struct_inherits_from(Grt.PGrt,
      ContentStructName, 'db.DatabaseObject')=1) then
    begin
      C := 0;

      //Collect all objects for the sourceObject list
      //Loop over all schemata
      for J:=0 to myx_grt_list_item_count(PSchemata)-1 do
      begin
        PSchema := myx_grt_list_item_get(PSchemata, J);

        PSchemaItemList :=
          myx_grt_dict_item_get_value(PSchema, MemberName);

        if (PSchemaItemList<>nil)then
        begin
          //Add objects to sourceObject list
          for K:=0 to myx_grt_list_item_count(PSchemaItemList)-1 do
          begin
            PSchemaItem := myx_grt_list_item_get(PSchemaItemList, K);

            myx_grt_list_item_add_as_string(PSourceObjects,
              myx_grt_dict_id_item_as_string(
                PSchemaItem));

            inc(C);
          end;
        end;
      end;

      //Create FilterFrame
      if (C > 0) then
      begin
        FilterFrame := TMigrationObjSelFilterFrame.Create(self, Grt);
        FFilterFrames.Add(FilterFrame);
        FilterFrame.Name := 'FilterFrame' + IntToStr(FFilterFrames.Count);
        FilterFrame.Parent := ObjSelScrollBox;

        FilterFrame.StructName := ContentStructName;
      end;
    end;
  end;

  for I:=0 to FFilterFrames.Count-1 do
    TMigrationObjSelFilterFrame(FFilterFrames[I]).RefreshLists;

  MainPageControl.ActivePage := IgnoreListSheet;

  NextBtn.Enabled := True;
  if NextBtn.CanFocus then
    NextBtn.SetFocus;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TReverseEngineeringForm.DoFinalizeRevEng;

var
  I, Space, Cols: Integer;
  SchemaList, Schema,
  Model: Pointer;
  CurrentView: Pointer;
  ViewElementList: Pointer;
  Task: IGrtTask;

begin
  MigrationImg.Picture.Assign(FTaskUncheckedPNGImg);
  AddObjectsImg.Picture.Assign(FTaskUncheckedPNGImg);
  PlaceObjectsImg.Picture.Assign(FTaskUncheckedPNGImg);

  // hide migration if it is not needed
  MigrationLbl.Visible := False;
  MigrationImg.Visible := False;


  ResultFinalizeLbl.Caption := '';
  ResultFinalizeLbl.Visible := True;

  // Flip page
  MainPageControl.ActivePage := FinalizeSheet;
  UpdateAdvancedBtnState;

  DoProcessStart;
  ResultFinalizeLbl.Caption := _('Removing ignored object from catalog');

  Application.ProcessMessages;

  //Filter out objects from ignore list
  Task := Grt.CreateStandardTask(
    ResultFinalizeLbl.Caption,
    'DbUtils',
    'removeIgnoredObjectsFromCatalog',
    [Grt.Global['/migration/sourceCatalog'],
      Grt.Global['/migration/ignoreList']],
    ProcessGrtOutput,
    ProcessGrtMessages,
    True,
    False,
    -1,
    PerformGrtInput, nil, TaskFinished);
  Grt.AddTask(Task);

  // remove empty catalogs
  Task := Grt.CreateStandardTask(
    _('Removing empty schemata'),
    'DbUtils',
    'removeEmptySchemataFromCatalog',
    [Grt.Global['/workbench/catalog']],
    ProcessGrtOutput,
    ProcessGrtMessages,
    True,
    False,
    -1,
    PerformGrtInput, nil, TaskFinished);
  Grt.AddTask(Task);

  // Merge catalogs
  FProcessCatalogsTask := Grt.CreateStandardTask(
    _('Merging catalogs'), 'DbUtils', 'mergeCatalogs', [Grt.Global['/migration/sourceCatalog'],
    Grt.Global['/workbench/catalog']], ProcessGrtOutput, ProcessGrtMessages, True, False, -1, PerformGrtInput,
    nil, TaskFinished
  );
  Grt.AddTask(FProcessCatalogsTask);

  // Finally place the objects on the canvas.
  if PlaceAlgoComboBox.ItemIndex > 0 then
  begin
    ResultFinalizeLbl.Caption := _('Place objects on the canvas');
    ResultFinalizeLbl.Update;

    Model := Grt.Global['/workbench/model'];
    SchemaList := Grt.Global['/migration/sourceCatalog/schemata'];
    for I := 0 to Grt.ListCount(SchemaList) - 1 do
    begin
      Schema := Grt.ListItem[SchemaList, I];

      // Check if the current view is empty.
      CurrentView := Grt.DictRef[Model, 'currentView'];
      ViewElementList := Grt.DictItem[CurrentView, 'elements'];
      if (Grt.ListCount(ViewElementList) > 0) then
      begin
        // if not, add a new view
        Task := Grt.CreateStandardTask(
          _('Adding new view'),
          'Workbench',
          'addView',
          [Model, Grt.DictString[Schema, 'name']],
          ProcessGrtOutput,
          ProcessGrtMessages,
          True,
          False,
          -1,
          PerformGrtInput
        );
        Grt.AddTask(Task);  
      end;

      case PlaceAlgoComboBox.ItemIndex of
        1: // Automatic arrangement.
          begin
            FArrangeTask := Grt.CreateStandardTask(
              _('Arranging schema objects on current view'), 'Workbench', 'loadAndAutoArrange',
              [Schema, CurrentView], ProcessGrtOutput, ProcessGrtMessages, True, False, -1,
              PerformGrtInput, nil, TaskFinished);
            Grt.AddTask(FArrangeTask);
          end;
        2: // Grid arrangement.
          begin
            case PlaceAreaComboBox.ItemIndex of
              0:
                Cols := 5;
              1:
                Cols := 7;
              2:
                Cols := 10;
              3:
                Cols := 15;
              4:
                Cols := 20;
              5:
                Cols := 25;
            else
              Cols := 70;
            end;


            case PlaceSpaceComboBox.ItemIndex of
              0:
                Space := 40;
              1:
                Space := 70;
              2:
                Space := 100;
            else
              Space := 70;
            end;

            FArrangeTask := Grt.CreateStandardTask(
              _('Arranging schema objects on current view'), 'Workbench', 'arrangeSchemaOnCurrentView',
              [Schema, 100, 100, Cols, Space], ProcessGrtOutput, ProcessGrtMessages, True, False, -1,
              PerformGrtInput, nil, TaskFinished);
            Grt.AddTask(FArrangeTask);
          end;
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TReverseEngineeringForm.PerformGrtInput(Description: WideString): WideString;

var
  InputText: WideString;

begin
  InputText := '';

  ShowModalEditDialog(_('Input'),
    Description, myx_mtEdit,
    _('Ok'), True, '', InputText);

  Result := InputText;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TReverseEngineeringForm.PlaceAlgoComboBoxCloseUp(Sender: TObject);

begin
  PlaceAreaLbl.Enabled := (PlaceAlgoComboBox.ItemIndex > 1);
  PlaceAreaComboBox.Enabled := PlaceAreaLbl.Enabled;

  PlaceSpaceLbl.Enabled := PlaceAreaLbl.Enabled;
  PlaceSpaceComboBox.Enabled := PlaceAreaLbl.Enabled;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TReverseEngineeringForm.TaskFinished(Task: IGrtGenericTask);

// Called when an asynchronous task has been finished.

begin
  if Task = FFetchSchemaTask then
  begin
    if not TaskHandleError(Task) then
    begin
      FSchemaList := FFetchSchemaTask.Result;
      ConnImg.Picture.Assign(FTaskCheckedPNGImg);
      SchemaListImg.Picture.Assign(FTaskCheckedPNGImg);
    end;
    ResultLbl.Hide;
    FFetchSchemaTask := nil;
    DoProcessEnd;
  end
  else
    if Task = FReverseEngineerTask then
    begin
      if not TaskHandleError(Task) then
      begin
        FCatalog := FReverseEngineerTask.Result;
        Grt.Global['/migration/sourceCatalog'] := FCatalog;
        RevEngImg.Picture.Assign(FTaskCheckedPNGImg);
        DoBuildObjectSelection;
        CheckResImg.Picture.Assign(FTaskCheckedPNGImg);
      end;
      DoProcessEnd;
      Grt.ValueRelease(FSelectedSchemaList);
      FSelectedSchemaList := nil;
      Progress2Lbl.Hide;

      FReverseEngineerTask := nil;
    end
    else
      if Task = FProcessCatalogsTask then
      begin
        // Just show some feedback about the finished task.
        if not TaskHandleError(Task) then
          AddObjectsImg.Picture.Assign(FTaskCheckedPNGImg);
        FProcessCatalogsTask := nil;
      end
      else
        if Task = FArrangeTask then
        begin
          if not TaskHandleError(Task) then
          begin
            PlaceObjectsImg.Picture.Assign(FTaskCheckedPNGImg);
            PlaceObjectsLbl.Caption := _('Reverse engineering completed successfully.');
            NextBtn.Caption := _('Finish');
            BackBtn.Enabled := False;
            CancelBtn.Enabled := False;
          end
          else
          begin
            NextBtn.Enabled := False;
            BackBtn.Enabled := True;
            CancelBtn.Enabled := True;
          end;
          DoProcessEnd;
        end
        else
          TaskHandleError(Task);
end;

//----------------------------------------------------------------------------------------------------------------------

function TReverseEngineeringForm.TaskHandleError(Task: IGrtGenericTask): Boolean;

// Checks if the task has returned an error. If so the error message is shown, the error images are set and
// True is returned. Otherwise False is the result.

var
  StandardTask: IGrtTask;
  Memo: TTntMemo;
  
begin
  Result := False;

  if (Task.ErrorCode <> MYX_GRT_NO_ERROR) or (Task.ErrorString <> '') then
  begin
    Result := True;

    if (MainPageControl.ActivePage = GetSchemataSheet) then
      Memo := GRTMessageMemo
    else
      if (MainPageControl.ActivePage = RevEngSheet) then
        Memo := GRTMessage2Memo
      else
        if (MainPageControl.ActivePage = FinalizeSheet) then
          Memo := GRTMessage3Memo
        else
          Memo := nil;

    if Assigned(Memo) then
    begin
      if Supports(Task, IGrtTask, StandardTask) then
        Memo.Lines.Add(Format(
          _('A problem occured while executing task: %s. '#1310'The error occured in %s.%s and is: %s (code: %d).'),
          [StandardTask.Description, StandardTask.ModuleName, StandardTask.FunctionName, Task.ErrorString, Ord(Task.ErrorCode)]))
      else
        Memo.Lines.Add(Format(
          _('A problem occured while executing a shell task. '#1310'The error is: %s (code: %d).'),
          [Task.ErrorString, Ord(Task.ErrorCode)]));
    end;
    
    ConnImg.Picture.Assign(FTaskErrorPNGImg);
    SchemaListImg.Picture.Assign(FTaskErrorPNGImg);
    NextBtn.Enabled := False;
    BackBtn.Enabled := True;
    CancelBtn.Enabled := True;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TReverseEngineeringForm.SchemaListViewDblClick(Sender: TObject);

begin
  UpdateSchemaSelectionCountLbl;

  if SchemaListView.SelCount > 0 then
    DoRevEng;
end;

//----------------------------------------------------------------------------------------------------------------------

end.
